rawset(_G, 'hooks', {})

rawset(_G, 'addPTHook', function(name, func)
	if not hooks[name]
		hooks[name] = {}
	end
	table.insert(hooks[name], func)
end)

rawset(_G, 'executePTHook', function(name, table)
	if hooks[name]
		local return_val = nil
		for _,func in pairs(hooks[name])
			local thing = func(unpack(table))
			if not return_val
				return_val = thing
			end
		end
		return return_val
	end
end)

PizzaTime.functions.inits.PlayerInit = function(player)
	player.pt_mode = {
		lap = 1,
		pizzaface = false,
		pizzaface_stun = 30 * TICRATE,
		rank = #PizzaTime.sync.ranks,
		can_lap = false,
		pizzamask = nil,
		defender = false,
		extreme = false,
		leaving_stage = false,
		was_hit = false,
		specForce = false,
		hud = {
			rank_tween = 0
		},
		war_timer = {
			activated = false,
			time = 30 * TICRATE,
			saved_rings = 0
		},
		shotgun = {
			mobj = nil,
			pressed = false,
			can_use = false,
			stomp = false,
			stomp_speed = 0,
			cooldown = 0
		}
	}
	if not old_war_timer.value then
		player.pt_mode.war_timer.time = war_timer_time.value*TICRATE
	end
	if PizzaTime.sync.Defined then
		player.pt_mode.specForce = PizzaTime.sync.PizzaTime
	end
	executePTHook('PlayerInit', {player})
end

PizzaTime.functions.inits.MapInit = function(map)
	PizzaTime.sync.PizzaTime = false
	PizzaTime.sync.ExtremeLap = false
	PizzaTime.sync.Global_Laps = 1
	PizzaTime.sync.SpawnSector = nil
	PizzaTime.sync.SpawnThing = nil
	PizzaTime.sync.PizzaTimeSign = nil
	PizzaTime.sync.LaughedLol = false
	PizzaTime.sync.ShotgunGained = false
	PizzaTime.sync.Practice = false
	PizzaTime.sync.TitleCard = {
		time = 3*TICRATE,
		alpha = 10,
		alpha2 = 0,
	}
	PizzaTime.sync.TitleCardBool = false
	PizzaTime.sync.TitleCardName = nil
	if PizzaTime.sync.maps_with_titlecards[G_BuildMapTitle(map)] then
		PizzaTime.sync.TitleCardName = PizzaTime.sync.maps_with_titlecards[G_BuildMapTitle(map)]
	end
	PizzaTime.sync.ranks[0].laps = max_laps.value+max_extreme_laps.value
	PizzaTime.sync.LapPortalExists = false
	PizzaTime.sync.SpecialPlayers.PizzaFaces = {}
	PizzaTime.sync.SpecialPlayers.ExtremeLapper = {}
	PizzaTime.sync.Pizzaface_AI = {}
	PizzaTime.sync.Pizzaface_AI.mobj = nil
	PizzaTime.sync.Pizzaface_AI.pizzaface_stun = 30*TICRATE
	PizzaTime.sync.hud.PizzaTimeY = 0
	PizzaTime.sync.hud.PizzaTimeFlash = 1
	PizzaTime.sync.hud.PizzaTimeTimerY = 0
	PizzaTime.sync.hud.IntermissionSongTimer = 15*TICRATE
	PizzaTime.sync.hud.IntermissionSong = false
	PizzaTime.sync.hud.OrganizedSP = false
	PizzaTime.sync.hud.Showtime = 0
	PizzaTime.sync.hud.Lap2Objs = {}
	PizzaTime.sync.hud.SavedPlayers = {}
	animation_handler = {}
	executePTHook('DefineVars', {})

	INIT_Animation('john_timer',0,0,22,2,'JOHN',true)
	INIT_Animation('pizzaface_isleep',0,0,17,2,'PFCES',true)

	for player in players.iterate
		player.pt_mode = nil
	end

	if gametype == GT_PIZZATIME2
		hud.disable("intermissionmessages")
		hud.disable("intermissiontally")
	else
		hud.enable("intermissionmessages")
		hud.enable("intermissiontally")
	end

	executePTHook('MapInit', {})
	PizzaTime.sync.Defined = true
end

PizzaTime.functions.inits.MapFinishInit = function()
	if gametype ~= GT_PIZZATIME2 then return end
			
	for thing in mapthings.iterate
		if thing.type == 501
			local john = nil
			if not executePTHook('Before_SpawnPillarJohn', {thing.x,thing.y,thing.z})
				PizzaTime.sync.EndThing = thing
				john = P_SpawnMobj(thing.x*FU,thing.y*FU,thing.z*FU,MT_PILLARJOHN)
			end
			executePTHook('SpawnPillarJohn', {thing.x,thing.y,thing.z,john})
		end
		if thing.type == 1
			PizzaTime.sync.SpawnThing = thing
		end
		if thing.type == 2048 then
			PizzaTime.sync.LapPortalExists = true
		end
	end
end

PizzaTime.functions.hooks.PizzaTimeMusicThinker = function(old, new)
	if gametype ~= GT_PIZZATIME2 then return end
	if new == '_inter' then
		return true
	end
	local titlecard = PizzaTime.sync.TitleCardName
	if titlecard and new == titlecard[1] and not PizzaTime.sync.TitleCardBool then
		return true
	end
end

local saved_end_coords = {}

addHook('NetVars', function(n)
	saved_end_coords = n($)
end)

addHook('MapChange', function()
	saved_end_coords = {}
end)

PizzaTime.functions.tools.TPToEnd = function(mo, tp_mo)
	if PizzaTime.sync.EndThing
		if not executePTHook('Before_TPToEnd', {mo, tp_mo})
			P_SetOrigin(mo, PizzaTime.sync.EndThing.x*FU,PizzaTime.sync.EndThing.y*FU,PizzaTime.sync.EndThing.z*FU)
			mo.angle = PizzaTime.sync.EndThing.angle
		end
		executePTHook('TPToEnd', {mo, tp_mo})
	else
		if not saved_end_coords[1]
			saved_end_coords = {tp_mo.x, tp_mo.y, tp_mo.z}
		else
			if not executePTHook('TPToEnd', {mo, tp_mo, saved_end_coords})
				P_SetOrigin(mo, saved_end_coords[1], saved_end_coords[2], saved_end_coords[3])
			end
			executePTHook('TPToEnd', {mo, tp_mo, saved_end_coords})
		end
	end
end

PizzaTime.functions.tools.GetNearestPlayer = function(mo)
	local nearest_player = nil
	local nearest_coords = {x=nil,y=nil}
	for p in players.iterate
		if p.mo and p.mo.valid
			local x,y = abs(mo.x-p.mo.x),abs(mo.y-p.mo.y)
			if (not nearest_coords.x or not nearest_coords.y) or (x < nearest_coords.x and y < nearest_coords.y)
				nearest_coords = {}
				nearest_coords.x = x
				nearest_coords.y = y
				nearest_player = p
			end
		end
	end
	return nearest_player
end

PizzaTime.functions.triggers.TriggerPT = function(p)
	if not PizzaTime.sync.PizzaTime
		local exiting,total = PizzaTime.functions.tools.PlayerCounter()
		if not executePTHook('Before_PizzaTimeStart', {p, exiting, total})
		  PizzaTime.sync.PizzaTime = true
	  		if total > 1
				p.pt_mode.pizzaface = true
	  			table.insert(PizzaTime.sync.SpecialPlayers.PizzaFaces, p)
	  		else
				PizzaTime.functions.tools.TPToEnd(p.mo, p.mo) // saves the damn coords so it doesnt fuck itself up
				PizzaTime.sync.Practice = true
				if PizzaTime.sync.EndThing
					PizzaTime.functions.triggers.SpawnPizzaface(PizzaTime.sync.EndThing.x*FU, PizzaTime.sync.EndThing.y*FU, PizzaTime.sync.EndThing.z*FU, p)
				else
					PizzaTime.functions.triggers.SpawnPizzaface(p.mo.x,p.mo.y,p.mo.z,p)
				end
	  		end
	  		for player in players.iterate
	  			if player.mo and player.mo.valid and player != p
	  				PizzaTime.functions.tools.TPToEnd(player.mo, p.mo)
	  			end
	  		end
	  		PizzaTime.sync.PizzaTimeSign = P_SpawnMobj(PizzaTime.sync.SpawnThing.x*FRACUNIT,PizzaTime.sync.SpawnThing.y*FRACUNIT,PizzaTime.sync.SpawnThing.z*FRACUNIT,MT_SIGN)
	  		if PizzaTime.sync.SpawnThing.angle
	  			PizzaTime.sync.PizzaTimeSign.angle = PizzaTime.sync.SpawnThing.angle
	  		end
	  		if PizzaTime.sync.PizzaTimeSign.subsector.sector
	  			PizzaTime.sync.EndSector = PizzaTime.sync.PizzaTimeSign.subsector.sector
	  		end
			executePTHook('PizzaTimeStart', {p})
	  	end
	end
end

PizzaTime.functions.checks.CanLap = function(p)
	local pt = p.pt_mode
	local maxext = max_laps.value+max_extreme_laps.value
	if pt.lap < max_laps.value then
		return true,false
	end
	if lap_3.value and PizzaTime.sync.ExtremeLap == pt.extreme then
		if pt.lap >= max_laps.value and pt.lap < maxext then
			if not max_extreme_rings.value then
				return true,true
			end
			
			if p.rings >= max_extreme_rings.value then
				return true,true
			end
		end
	end
	return false,false
end

PizzaTime.functions.checks.PizzaPlayer = function(p)
	if p.exiting then
		if not lap_3_exiting_players_become_pizzaface.value then
			return false
		end
	end
	return true
end

PizzaTime.functions.triggers.NewLap = function(p)
	if p.pt_mode and p.mo and p.mo.valid and p.pt_mode.can_lap
		local pt = p.pt_mode
		local canlap,ext = PizzaTime.functions.checks.CanLap(p)

		if canlap then
			if not executePTHook('Before_NewLap', {p,ext})
	  			P_AddPlayerScore(p, 500)
	  			pt.lap = $+1
	  			PizzaTime.functions.tools.TPToEnd(p.mo)
				if PizzaTime.sync.Pizzaface_AI.mobj and PizzaTime.sync.Pizzaface_AI.mobj.valid then
					PizzaTime.functions.triggers.SpawnPizzaface(p.mo.x,p.mo.y,p.mo.z,p)
				end
	  			p.exiting = 0
	  			if pt.lap > PizzaTime.sync.Global_Laps
	  				PizzaTime.sync.Global_Laps = pt.lap
	  			end
	  			if ext and not PizzaTime.sync.ExtremeLap
					if not executePTHook('Before_ExtremeToggle', {p})
						PizzaTime.sync.ExtremeLap = true
						p.pt_mode.extreme = true
						pt.war_timer.activated = true
						pt.shotgun.can_use = true
						PizzaTime.sync.SpecialPlayers.ExtremeLapper[1] = p
						P_SetupLevelSky(34)
						if lap_3_multi_pizzafaces.value
							for player in players.iterate
								if player == p then continue end
								if player.mo and player.mo.valid and PizzaTime.functions.checks.PizzaPlayer(player)
									if not player.pt_mode.pizzaface and not executePTHook('Extreme_BecomePizzaface', {p, player})
										player.pt_mode.pizzaface = true
										table.insert(PizzaTime.sync.SpecialPlayers.PizzaFaces, player)
									end
									player.pt_mode.pizzaface_stun = 2*TICRATE
									PizzaTime.functions.tools.TPToEnd(player.mo)
								end
							end
						end
						executePTHook('ExtremeToggle', {p})
					end
	  			end
	    		newPlayerLapping(p.name, p.pt_mode.lap, ext)
	  		end
			executePTHook('NewLap', {p,ext})
		end
	end
end

PizzaTime.functions.hooks.ExitedPlayerThinker = function(p)
	if not PizzaTime.sync.Defined then return end
	local pt = p.pt_mode
	if not p.exiting then return end
	if pt.pizzaface then return end
	if p.mo and p.mo.valid and pt and pt.can_lap
		if PizzaTime.functions.checks.CanLap(p) and not PizzaTime.sync.LapPortalExists
			if p.cmd.buttons & BT_ATTACK
				PizzaTime.functions.triggers.NewLap(p)
			end
		end
	end
end

PizzaTime.functions.checks.PlayerSpotCheck = function(mo,line)
	if not mo.player then return end
	/*if ((line.frontsector and GetSecSpecial(line.frontsector.special, 4) == 2) 
    or (line.backsector and GetSecSpecial(line.backsector.special, 4) == 2))
	and not PizzaTime.sync.PizzaTime
		PizzaTime.functions.triggers.TriggerPT(mo.player)
    end*/

	-- Sign at start check.
	if PizzaTime.sync.EndSector
		if ((line.backsector and line.backsector == PizzaTime.sync.EndSector)
		or (line.frontsector and line.frontsector == PizzaTime.sync.EndSector))
		and not mo.player.exiting then
			if not mo.player.pt_mode.pizzaface then
				if not executePTHook('Before_PlayerFinish', {mo,true}) then
					if (mo.player.pt_mode.lap >= max_laps.value+max_extreme_laps.value or mo.player.pt_mode.defender) and defender.value then
						if not PizzaTime.sync.ShotgunGained then
							PizzaTime.sync.ShotgunGained = true
							mo.player.pt_mode.defender = true
							mo.player.pt_mode.war_timer.activated = false
							mo.player.pt_mode.lap = $+1
							PizzaTime.functions.tools.TPToEnd(mo)
						end
						if mo.player.pt_mode.defender then
							PizzaTime.functions.tools.TPToEnd(mo)
						end
						mo.player.pt_mode.can_lap = false
					else
						mo.player.pt_mode.can_lap = true
						P_DoPlayerExit(mo.player)
					end
	  				executePTHook('PlayerFinish', {mo,true})
				end
			else
				if not executePTHook('Before_PizzaFaceFinish', {mo,true}) then
					PizzaTime.functions.tools.TPToEnd(mo)
					mo.player.pt_mode.pizzaface_stun = 2*TICRATE
					executePTHook('PizzaFaceFinish', {mo,true})
				end
			end
		end
	end
end

PizzaTime.functions.hooks.PlayerSpawnFunc = function(p)
	if gametype ~= GT_PIZZATIME2 return end
	if not executePTHook('PlayerSpawnFunc', {p})
		if PizzaTime.sync.PizzaTime
			PizzaTime.functions.tools.TPToEnd(p.mo)
		end
		if p.pt_mode and p.pt_mode.pizzaface and (not p.pt_mode.pizzamask or (p.pt_mode.pizzamask and not p.pt_mode.pizzamask.valid))
			local pt = p.pt_mode
			pt.pizzamask = P_SpawnMobj(p.realmo.x,p.realmo.y,p.realmo.z,MT_PIZZAMASK)
			pt.pizzamask.targetplayer = p --dream reference
			pt.pizzamask.scale = FU/2
		end
		if p.pt_mode and p.pt_mode.shotgun.can_use and (not p.pt_mode.shotgun.mobj or (p.pt_mode.shotgun.mobj and not p.pt_mode.shotgun.mobj.valid))
			shotgun.mobj = P_SpawnMobjFromMobj(p.mo, (16*FRACUNIT)*cos(p.mo.angle), (16*FRACUNIT)*sin(p.mo.angle), p.mo.height/2, MT_SHOTGUN)
			shotgun.mobj.linked_mo = p.mo
		end
	end
end

PizzaTime.functions.checks.CanGetRank = function(rank, score, rings, laps, hit)
	if PizzaTime.sync.ranks[rank] then
		local curRank = PizzaTime.sync.ranks[rank]
		if score >= curRank.score then
			if not executePTHook('DetermineRank', {score, rings, laps, hit, 'score'})
				if not curRank.rings then
					return true
				end

				if rings >= curRank.rings then
					if not executePTHook('DetermineRank', {score, rings, laps, hit, 'rings'})
						if not curRank.laps then
							return true
						end

						if laps >= curRank.laps
							if not executePTHook('DetermineRank', {score, rings, laps, hit, 'laps'})
								if not curRank.cantbehit then
									return true
								end

								if not hit then
									if not executePTHook('DetermineRank', {score, rings, laps, hit, 'hit'})
										return true
									end
								end
							end
						end
					end
				end
			end
		end
		return false
	end
end

PizzaTime.functions.checks.UnderRank = function(rank, score, rings, laps, hit)
	if PizzaTime.sync.ranks[rank] then
		local curRank = PizzaTime.sync.ranks[rank]
		if score < curRank.score and not executePTHook('DetermineLowerRank', {score, rings, laps, hit, 'score'}) then
			return true
		end
		if not curRank.rings then return false end
		if rings < curRank.rings and not executePTHook('DetermineLowerRank', {score, rings, laps, hit, 'rings'}) then
			return true
		end
		if not curRank.laps then return false end
		if laps < curRank.laps and not executePTHook('DetermineLowerRank', {score, rings, laps, hit, 'laps'}) then
			return true
		end
		if not curRank.cantbehit and not executePTHook('DetermineLowerRank', {score, rings, laps, hit, 'hit'}) then return false end
		if hit then
			return true
		end
		return false
	end
end

PizzaTime.functions.tools.PlayerCounter = function()
	local exitingCount = 0
	local playerCount = 0

	for p in players.iterate
		if p.valid
			if p.exiting or p.pt_mode.specForce or p.pt_mode.pizzaface or p.pt_mode.defender
				exitingCount = $+1
			end
		end
		playerCount = $+1
	end

	return exitingCount,playerCount
end

PizzaTime.functions.triggers.SpawnPizzaface = function(x,y,z,p)
	if not PizzaTime.sync.Pizzaface_AI.mobj
		if not executePTHook('SpawnAIPizzaface', {x,y,z,p})
			PizzaTime.sync.Pizzaface_AI.mobj = P_SpawnMobjFromMobj(p.mo, 0,0,0, MT_PIZZAFACE)
			PizzaTime.sync.Pizzaface_AI.mobj.gofor = p.mo
			PizzaTime.sync.Pizzaface_AI.mobj.speedToGo = (skins[p.mo.skin].normalspeed - (8 * FRACUNIT)) / FRACUNIT
		end
	elseif PizzaTime.sync.Pizzaface_AI.mobj and PizzaTime.sync.Pizzaface_AI.mobj.valid and not PizzaTime.sync.Pizzaface_AI.pizzaface_stun and p.mo and p.mo == PizzaTime.sync.Pizzaface_AI.mobj.gofor
		if not executePTHook('TeleportAIPizzaface', {x,y,z,p})
			PizzaTime.sync.Pizzaface_AI.pizzaface_stun = 10*TICRATE
			P_SetOrigin(PizzaTime.sync.Pizzaface_AI.mobj, p.mo.x,p.mo.y,p.mo.z)
		end
	end
end

PizzaTime.functions.tools.ExtremeLapOverlayDrawer = function(v)
	if not PizzaTime.sync.ExtremeLap then return end
	local thing = v.cachePatch('BLACKG')
	v.drawStretched(0,0,v.width()*FU,v.height()*FU,thing,V_SNAPTOLEFT|V_SNAPTOTOP|7<<V_ALPHASHIFT)
end

PizzaTime.functions.tools.TimerIncrease = function(p, increase)
	local pt = p.pt_mode
	local wart = pt.war_timer
	if not wart.activated then return end
	if not old_war_timer.value then return end
	local oldtime = wart.time
	if wart.activated then
		if not executePTHook('Before_TimerIncrease', {p, wart.time, increase})
			if (p.powers[pw_shield] == SH_ATTRACT)
				wart.time = $ + ((TICRATE/2) * increase)
			else
				wart.time = $ + (TICRATE * increase)
			end
		end
		local newtime = wart.time
		executePTHook('TimerIncrease', {p, oldtime, newtime, increase})
	end
end

PizzaTime.functions.tools.CollectRings = function(ring, i, s)
	if gametype ~= GT_PIZZATIME2 return end
	if not s then return end
	if not s.player then return end
	P_AddPlayerScore(s.player, 15)
	PizzaTime.functions.tools.TimerIncrease(s.player, 1)
end

PizzaTime.functions.tools.KillRingBox = function(ring, i, s)
	if gametype ~= GT_PIZZATIME2 return end
	if not s then return end
	if not s.player then return end
	P_AddPlayerScore(s.player, 150)
	PizzaTime.functions.tools.TimerIncrease(s.player, 10)
end

-- hook time :)

PizzaTime.functions.hooks.PlayerThinker = function(p)
	if not p.mo then return end
	if not p.mo.valid then return end
	local pt = p.pt_mode
	if pt.specForce and not p.spectator
		p.cmd.forwardmove = 0
		p.cmd.sidemove = 0
		p.spectator = true
		P_RemoveShield(p)
		if pt.shotgun.mobj then
			P_RemoveMobj(pt.shotgun.mobj)
		end
	end

	if PizzaTime.sync.TitleCardName and PizzaTime.sync.TitleCard.time then
		p.pflags = $|PF_FULLSTASIS
		p.realtime = 0
	end

	if p.mo and p.mo.valid then
		if (p.mo.state == S_PLAY_PAIN or p.mo.state == S_PLAY_DEAD) and not p.pt_mode.was_hit then
			p.pt_mode.was_hit = true
		end
		local rank = PizzaTime.sync.ranks[pt.rank]
		local nextRank = PizzaTime.sync.ranks[pt.rank - 1]
		local prevRank = PizzaTime.sync.ranks[pt.rank + 1]

		if nextRank and PizzaTime.functions.checks.CanGetRank(pt.rank-1, p.score, p.rings, p.pt_mode.lap, p.pt_mode.was_hit) then
			pt.rank = $-1
			pt.hud.rank_tween = FRACUNIT/2
		end
		if rank and prevRank and PizzaTime.functions.checks.UnderRank(pt.rank, p.score, p.rings, p.pt_mode.lap, p.pt_mode.was_hit)
			pt.rank = $+1
			pt.hud.rank_tween = FRACUNIT/2
		end
	end
end

PizzaTime.functions.hooks.PizzafaceThinker = function(player) -- alot of the code here is by jisk
	if player.realmo and player.realmo.valid and player.pt_mode.pizzaface then
		local pt = player.pt_mode
		P_SetScale(player.realmo, 2*FU)
		player.realmo.flags2 = $|MF2_DONTDRAW -- invisible so that the pizza mask can draw over.
		player.outofcoop = false 
		if not pt.pizzamask then
			pt.pizzamask = P_SpawnMobj(player.realmo.x,player.realmo.y,player.realmo.z,MT_PIZZAMASK)
			pt.pizzamask.targetplayer = player --dream reference
			pt.pizzamask.scale = FU/2
		end

		if pt.pizzaface_stun then -- player freeze decrement (mainly for pizza faces)
			pt.pizzaface_stun = $-1
			player.pflags = $|PF_FULLSTASIS
		end
		/*
		if not p.realmo.colorized or p.realmo.color != SKINCOLOR_ORANGE
			p.realmo.color = SKINCOLOR_ORANGE
			p.realmo.colorized = true		
		end
		*/
		if pt.pizzaface_stun == 0 and not PizzaTime.sync.LaughedLol
			S_StartSound(nil, sfx_pizzah)
			PizzaTime.sync.LaughedLol = true
			INIT_Animation('pizzaface_showtime',0,0,8,3,'PFCEA',false)
		end
	end
end

local function fire(p, ang)
	if not executePTHook('Before_ShotgunFire', {p})
		for i = 1,3
			local angles = {0 - ANG10, 0, ANG10}
			local angletogo = p.mo.angle + angles[i]
			local aimingtogo = ang or p.aiming
			local bullet = P_SPMAngle(p.mo, MT_SHOTGUN_BULLET, angletogo, aimingtogo, MF2_SCATTER)
			if bullet then
				bullet.momx = 70*cos(bullet.angle)
				bullet.momy = 70*sin(bullet.angle)
				bullet.momz = 70*sin(aimingtogo)
			end
		end
		S_StartSound(p.mo, sfx_shosho)
		executePTHook('ShotgunFire', {p})
	end
end

PizzaTime.functions.hooks.PlayerShotgunThinker = function(p)
	local pt = p.pt_mode
	local shotgun = pt.shotgun

	if shotgun.can_use then
		if shotgun.mobj == nil
			shotgun.mobj = P_SpawnMobjFromMobj(p.mo, (16*FRACUNIT)*cos(p.mo.angle), (16*FRACUNIT)*sin(p.mo.angle), p.mo.height/2, MT_SHOTGUN)
			shotgun.mobj.linked_mo = p.mo
		end
		if (p.cmd.buttons & BT_ATTACK or p.cmd.buttons & BT_SPIN) and not shotgun.pressed and not shotgun.cooldown and not P_PlayerInPain(p) and p.playerstate == PST_LIVE and p.exiting == 0
			shotgun.pressed = true
			if not shotgun.stomp then
				if not (p.cmd.buttons & BT_SPIN)
					fire(p)
					shotgun.cooldown = TICRATE/4
				elseif p.cmd.buttons & BT_SPIN and not P_IsObjectOnGround(p.mo)
					local ang = -ANGLE_90
					fire(p, ang)
					shotgun.cooldown = TICRATE/4
					if p.pflags & PF_JUMPED
						if not executePTHook('Before_ShotgunStompInit', {p})
							p.mo.momz = 16*FRACUNIT
							shotgun.stomp = true
							p.pflags = $ & ~PF_JUMPED
							p.pflags = $ & ~PF_SPINNING
							p.pflags = $|PF_THOKKED
							p.mo.state = S_PLAY_FALL
							executePTHook('ShotgunStompInit', {p})
						end
					end
				end
			end
		elseif not (p.cmd.buttons & BT_ATTACK) and not (p.cmd.buttons & BT_SPIN) and shotgun.pressed
			shotgun.pressed = false
			S_StartSound(p.mo, sfx_sholod)
		end
		if shotgun.stomp
			if not P_IsObjectOnGround(p.mo)
				p.mo.momz = $ -(((FRACUNIT * 4) / 3) * P_MobjFlip(p.mo))
				shotgun.stomp_speed = p.mo.momz
			else
				if not executePTHook('Before_ShotgunStompEnd', {p})
					shotgun.stomp = false
					local sector = p.mo.subsector.sector
					if (p.cmd.buttons & BT_SPIN) and (sector.f_slope or sector.c_slope)
						local slope = sector.f_slope
						if not slope then return end
						p.pflags = $|PF_SPINNING
						p.mo.state = S_PLAY_ROLL
						local ang = 0
						if slope.zdelta < 0
							ang = ANGLE_180
						end
						P_Thrust(p.mo, slope.xydirection + ang, abs(shotgun.stomp_speed))
						p.mo.angle = slope.xydirection + ang
						S_StartSound(p.mo, sfx_zoom)
					else
						if shotgun.stomp_speed/FRACUNIT <= -32
							for i = 0,16 do
								P_SpawnSkidDust(p, 28*FRACUNIT)
							end
							S_StartSound(p.mo, sfx_s3k4c)
						end
					end
					shotgun.stomp_speed = 0
					executePTHook('ShotgunStompEnd', {p})
				end
			end
		end
		if shotgun.cooldown > 0
			shotgun.cooldown = $-1
		end
	else
		if shotgun.mobj then
			P_KillMobj(shotgun.mobj)
			shotgun.mobj = nil
		end
	end
end

PizzaTime.functions.hooks.PizzafaceKill = function(mo,mo2)
	if mo.player.pt_mode.pizzaface_stun == 0 then
		if mo2.player and not mo2.player.pt_mode.pizzaface and not mo2.player.pt_mode.specForce and not mo2.player.exiting then
			if not executePTHook('Before_PizzaFaceDeath')
				chatprint(' * '..mo2.player.name..' has died.')
				mo2.player.pt_mode.specForce = true
			end
			executePTHook('PizzaFaceDeath', {mo2.player, mo.player})
		end
	end
end

PizzaTime.functions.hooks.WarThinker = function(p)
	local wart = p.pt_mode.war_timer
	if not wart.activated then return end

	if not p.pt_mode.specForce and not p.exiting and p.mo and p.mo.valid
		wart.time = $-1
		if wart.time <= 0 * TICRATE
			if not executePTHook('Before_TimerDeath', {p})
				p.pt_mode.specForce = true
				chatprint(' * '..p.name..' has died due to the War timer.')
			end
			executePTHook('TimerDeath', {p})
		end

		if wart.time > war_timer_time.value * TICRATE
			if not executePTHook('Before_TimerCorrection', {p})
				wart.time = war_timer_time.value * TICRATE
			end
			executePTHook('TimerCorrection', {p})
		end
	end
end

PizzaTime.functions.hooks.HUDThinker = function()
	if PizzaTime.sync.PizzaTime
		if PizzaTime.sync.hud.PizzaTimeY < 500 * FU
			PizzaTime.sync.hud.PizzaTimeY = $+((FRACUNIT*6)/2)
		end
	
		if PizzaTime.sync.hud.PizzaTimeTimerY < 130 and not PizzaTime.sync.LaughedLol
			PizzaTime.sync.hud.PizzaTimeTimerY = $+1
		end

		if PizzaTime.sync.LaughedLol then
			PizzaTime.sync.hud.Showtime = $+1
			if PizzaTime.sync.hud.Showtime > TICRATE*3 and PizzaTime.sync.hud.PizzaTimeTimerY > 0
				PizzaTime.sync.hud.PizzaTimeTimerY = $-1
			end
		end

		if leveltime % 4 == 0
			if not executePTHook('HUD_OnPizzaTimeGrphFlash', {})
				if PizzaTime.sync.hud.PizzaTimeFlash == 1
					PizzaTime.sync.hud.PizzaTimeFlash = 2
				else
					PizzaTime.sync.hud.PizzaTimeFlash = 1
				end
			end
		end
	end

	for p in players.iterate do
		if p.pt_mode then
			local pt = p.pt_mode
			local hud = pt.hud
			if hud.rank_tween > 0 then
				local rank_tween = FRACUNIT/8
				if (hud.rank_tween - rank_tween) < 0 then
					hud.rank_tween = 0
				else
					hud.rank_tween = $-rank_tween
				end
			end
		end
	end

	for _,obj in pairs(PizzaTime.sync.hud.Lap2Objs)
		if not executePTHook('HUD_Before_LapObjThinker', {})
			if obj.y < 40 and not obj.goUp
				obj.y = $+1
			elseif obj.goUp
				obj.y = $-1
				if obj.y < -20
					table.remove(PizzaTime.sync.hud.Lap2Objs, _)
				end
			else
				obj.freezeTime = $+1
				if obj.freezeTime > 5*TICRATE
					obj.goUp = true
				end
			end
			executePTHook('HUD_LapObjThinker', {})
		end
	end
end

PizzaTime.functions.hooks.GlobalThinker = function()
	if PizzaTime.sync.TitleCardName and PizzaTime.sync.TitleCard.time then
		if PizzaTime.sync.TitleCard.time == 3*TICRATE then
			S_ChangeMusic(PizzaTime.sync.TitleCardName[2], false)
		end
		PizzaTime.sync.TitleCard.time = $-1
	elseif not PizzaTime.sync.TitleCardBool then
		PizzaTime.sync.TitleCardBool = true
		S_ChangeMusic(mapmusname)
	end
	if PizzaTime.sync.TitleCardName then
		if PizzaTime.sync.TitleCard.time > (3*TICRATE) - 10 and PizzaTime.sync.TitleCard.alpha > 0 then
			PizzaTime.sync.TitleCard.alpha = $-1
		end
		if PizzaTime.sync.TitleCard.time < 10 and PizzaTime.sync.TitleCard.alpha < 10 then
			PizzaTime.sync.TitleCard.alpha = $+1
		end
		if not PizzaTime.sync.TitleCard.time and PizzaTime.sync.TitleCard.alpha2 < 10 then
			PizzaTime.sync.TitleCard.alpha2 = $+1
		end
	end
end
PizzaTime.functions.hooks.IntermissionThinker = function()
	if PizzaTime.sync.hud.IntermissionSongTimer > 0 then
		PizzaTime.sync.hud.IntermissionSongTimer = $-1
	elseif not PizzaTime.sync.hud.IntermissionSong then
		PizzaTime.sync.hud.IntermissionSong = true
		S_ChangeMusic('PITINT')
	end
	if not PizzaTime.sync.hud.SavedPlayers[1] and not PizzaTime.sync.hud.OrganizedSP then
		PizzaTime.sync.hud.SavedPlayers = {}
		for p in players.iterate do
			if p.pt_mode and not p.pt_mode.specForce and not p.pt_mode.pizzaface then
				table.insert(PizzaTime.sync.hud.SavedPlayers, p)
			end
		end

		table.sort(PizzaTime.sync.hud.SavedPlayers, function(a,b)
			local pt_1 = a.pt_mode
			local pt_2 = b.pt_mode
			if pt_1.rank != pt_2.rank then
				if pt_1.rank < pt_2.rank then
					return true
				end
			else
				if a.score > b.score then
					return true
				end
			end
		end)
		if PizzaTime.sync.hud.SavedPlayers[1] then
			S_ChangeMusic(PizzaTime.sync.ranks[PizzaTime.sync.hud.SavedPlayers[1].pt_mode.rank].name..'RANK', false)
		else
			S_ChangeMusic('DRANK', false)
			S_StartSound(nil, sfx_pizzah)
		end
		PizzaTime.sync.hud.OrganizedSP = true
	end
end

-- what the john doin

PizzaTime.functions.hooks.JohnSpawn = function(john)
	john.inv = 0
	john.hits = 2
	john.isdead = false
	john.hitby = {}
	john.hittimer = 10*TICRATE
	john.deadtime = 5*TICRATE
end

PizzaTime.functions.hooks.JohnThinker = function(john)
	if not john.valid then return end
	if john.hittimer and john.hitby[1] then
		john.hittimer = $-1
	elseif not john.hittimer and john.hitby[1]
		for _,i in pairs(john.hitby) do
			if i.valid then
				S_ChangeMusic(mapmusname, true, i)
			end
		end
		john.hitby = {}
		john.state = S_PILLARJOHN
	end
	if john.inv then
		john.inv = $-1
	end
	if john.isdead then
		if john.deadtime then
			john.momz = $-(FU/2)
			john.deadtime = $-1
		else
			P_RemoveMobj(john)
		end
	end
end

PizzaTime.functions.hooks.JohnDeath = function(john, i, s)
	if s then
		john.momx = s.momx
		john.momy = s.momy
	end
	john.momz = 16*FU
	john.flags = $|MF_NOCLIP|MF_NOCLIPHEIGHT
	john.isdead = true
end

PizzaTime.functions.hooks.JohnCollideThinker = function(john, mo)
	if not mo.player then return true end
	if PizzaTime.sync.PizzaTime then return end
	local p = mo.player
	if P_PlayerCanDamage(p, john) and not john.inv then
		if john.hits
			if not executePTHook('Before_HitPillarJohn', {john, mo})
				if not L_TableContains(john.hitby, p) then
					table.insert(john.hitby, p)
				end
				S_StartSound(john, sfx_ouchie)
				S_ChangeMusic('PHOBIA', true, p)
				john.hits = $-1
				john.inv = TICRATE
				john.hittimer = 10*TICRATE
				if john.state ~= S_PILLARJOHN_UHOH then john.state = S_PILLARJOHN_UHOH end
			end
			executePTHook('HitPillarJohn', {john, mo})
		else
			if not executePTHook('KillPillarJohn', {john, mo})
				P_KillMobj(john, nil, mo)
				S_StartSound(john, sfx_ouchie)
				PizzaTime.functions.triggers.TriggerPT(p)
				PizzaTime.sync.TitleCard.alpha2 = 0
			end
			executePTHook('KillPillarJohn', {john, mo})
		end
		mo.momx = $*-1
		mo.momy = $*-1
		mo.momz = $*-1
	end
	return true
end

PizzaTime.functions.hooks.LapPortalCollideThinker = function(portal, mo)
	if not mo.player then return true end
	if not mo.player.pt_mode then return true end
	if not PizzaTime.sync.PizzaTime then return true end
	mo.player.pt_mode.can_lap = true
	if not (PizzaTime.functions.checks.CanLap(mo.player) or mo.player.pt_mode.pizzaface) then return true end
	if not mo.player.pt_mode.pizzaface
		if not executePTHook('Before_PlayerFinish', {mo,false}) then
			PizzaTime.functions.triggers.NewLap(mo.player)
			executePTHook('PlayerFinish', {mo,true})
		end
	else
		if not executePTHook('Before_PizzaFaceFinish', {mo,false}) then
			PizzaTime.functions.tools.TPToEnd(mo)
			mo.player.pt_mode.pizzaface_stun = 2*TICRATE
			executePTHook('PizzaFaceFinish', {mo,false})
		end
	end
	S_StartSound(mo, sfx_portal, mo.player)
	return true
end

PizzaTime.functions.hooks.MonitorDeath = function(monitor, i, s)
	if not s then return end
	if not s.player then return end
	if not (monitor.flags & MF_MONITOR) then return end
	P_AddPlayerScore(s.player, 50)
end