rawset(_G, 'L_ZCollide', function(mo1,mo2)
	if mo1.z > mo2.height+mo2.z then return false end
	if mo2.z > mo1.height+mo1.z then return false end
	return true
end)

local function R_WorldToScreen2(p, camera, target, v)

	local visible = true
	local first_person = not camera.chase
	local cam = first_person and p.mo or camera
	local hudwidth = 320*FRACUNIT
	local hudheight = (320*v.height()/v.width())*FRACUNIT

	local fov = ANGLE_90 -- Can this be fetched live instead of assumed?

	-- the "distance" the HUD plane is projected from the player
	local hud_distance = FixedDiv(hudwidth / 2, tan(fov/2))
	if not target.valid then visible = false end

	if not P_CheckSight(p.mo, target) then visible = false end

	-- how far away is the other player?
	local distance = R_PointToDist(target.x, target.y)

	--Angle between camera vector and target
	local hangdiff = R_PointToAngle2(cam.x, cam.y, target.x, target.y)
	local hangle = hangdiff - cam.angle

	--check if object is outside of our field of view
	--converting to fixed just to normalise things
	--e.g. this will convert 365° to 5° for us
	local fhanlge = AngleFixed(hangle)
	local fhfov = AngleFixed(fov/2)
	local f360 = AngleFixed(ANGLE_MAX)
	if fhanlge < f360 - fhfov and fhanlge > fhfov then
		visible = false
	end

	--figure out vertical angle
	local h = FixedHypot(cam.x-target.x, cam.y-target.y)
	local vangdiff = R_PointToAngle2(0, 0, target.z-cam.z+target.height+20*FRACUNIT, h) - ANGLE_90
	local vcangle = first_person and p.aiming or cam.aiming or 0
	local vangle = vcangle + vangdiff

	--again just check if we're outside the FOV
	local fvangle = AngleFixed(vangle)
	local fvfov = FixedMul(AngleFixed(fov), FRACUNIT*v.height()/v.width())
	if fvangle < f360 - fvfov and fvangle > fvfov then
		visible = false
	end

	local hpos = hudwidth/2 - FixedMul(hud_distance, tan(hangle))
	local vpos = hudheight/2 + FixedMul(hud_distance, tan(vangle))

	local ss = FixedDiv(160*FU, distance)

	return {x=hpos, y=vpos, scale=ss, onscreen=visible}
end

rawset(_G, "R_WorldToScreen2", R_WorldToScreen2)

local FIXED_TICRATE = TICRATE * FRACUNIT

rawset(_G, 'animation_handler', {})

addHook('NetVars', function(n)
	animation_handler = n($)
end)

rawset(_G, 'INIT_Animation', function(name, x, y, frames, tps, img, loop)
	local tb = {}
	tb.x = x
	tb.y = y
	tb.frames = frames
	tb.curframe = 1
	tb.tps = tps
	tb.anim = img
	tb.loop = loop
	rawset(animation_handler, name, tb)
end)

addHook('ThinkFrame', function()
	for _,anim in pairs(animation_handler)
		if leveltime % anim.tps == 0
			anim.anim_time = 0
			if anim.curframe < anim.frames
				anim.curframe = $+1
			elseif anim.loop
				anim.curframe = 1
			end
		end
	end
end)

local function GetInternalFontWidth(str, font, fixed)

	-- No string
	if not (str) then return 0 end

	local width = 0

	for i=1,#str do
		-- (Using patch width by the way)
		if (font == "STCFN") then -- default font
			width = $1+8
		elseif (font == "TNYFN") then
			width = $1+7
		elseif (font == "LTFNT") then
			width = $1+20
		elseif (font == "TTL") then
			width = $1+29
		elseif (font == "CRFNT" or font == "NTFNT") then -- TODO: Credit font centers wrongly
			width = $1+16
		elseif (font == "NTFNO") then
			width = $1+20
		elseif (font == 'PTFNT')
			width = $1+27
		else
			width = $1+30
		end
	end
	if fixed
		width = $ * FRACUNIT
	end
	return width
end


local function drawSuperText(v, x, y, str, parms)

	-- Scaling
	local scale = (parms and parms.scale) or 1*FRACUNIT
	local hscale = (parms and parms.hscale) or 0
	local vscale = (parms and parms.vscale) or 0
	local yscale = (8*(FRACUNIT-scale))
	-- Spacing
	local xspacing = (parms and parms.xspace) or 0 -- Default: 8
	local yspacing = (parms and parms.yspace) or 4
	-- Text Font
	local font = (parms and parms.font) or "STCFN"
	local uppercs = (parms and parms.uppercase) or false
	local align = (parms and parms.align) or nil
	local flags = (parms and parms.flags) or 0
	local fixed = (parms and parms.fixed) or false
	local color = (parms and parms.color) or nil

	if font == 'PTFNT'
		scale = $ - FRACUNIT/2 - FRACUNIT/4
	end

	-- Split our string into new lines from line-breaks
	local lines = {}

	for ls in str:gmatch("[^\r\n]+") do
		table.insert(lines, ls)
	end

	-- For each line, set some stuff up
	for seg=1,#lines do
		
		local line = lines[seg]
		-- Fixed Position
		local fx = x << FRACBITS
		local fy = y << FRACBITS
		if fixed
			fx = x
			fy = y
		end
		-- Offset position
		local off_x = 0
		local off_y = 0
		-- Current character & font patch (we assign later later instead of local each char)
		local char
		local charpatch

		-- Alignment options
		if (align) then
			-- TODO: not working correctly for CRFNT
			if (align == "center") then
				if not fixed
					fx = $1-FixedMul( (GetInternalFontWidth(line, font, fixed)/2), scale) << FRACBITS -- accs for scale
				else
					fx = $1-FixedMul( (GetInternalFontWidth(line, font, fixed)/2), scale)
				end
				-- 	fx = $1-FixedMul( (v.stringWidth(line, 0, "normal")/2), scale) << FRACBITS
			elseif (align == "right") then
				if not fixed
					fx = $1-FixedMul( (GetInternalFontWidth(line, font, fixed)), scale) << FRACBITS
				else
					fx = $1-FixedMul( (GetInternalFontWidth(line, font, fixed)), scale)
				end
				-- fx = $1-FixedMul( (v.stringWidth(line, 0, "normal")), scale) << FRACBITS
			end
		end

		-- Go over each character in the line
		local charpatch = nil
		for strpos=1,#line do
			local drawable = true
			-- get our character step by step
			char = line:sub(strpos, strpos)

			-- TODO: custom skincolors will make a mess of this since the charlimit is 255
			-- Set text color, inputs, and more through special characters
			-- Referencing skincolors https://wiki.srb2.org/wiki/List_of_skin_colors

			-- TODO: effects?
			-- if (char:byte() == 161) then
			-- 	continue
			-- end
			-- print(strpos<<27)
			-- off_x = (cos(v.RandomRange(ANG1, ANG10)*leveltime))
			-- off_y = (sin(v.RandomRange(ANG1, ANG10)*leveltime))
			-- local step = strpos%3+1
			-- print(step)
			-- off_x = cos(ANG10*leveltime)*step
			-- off_y = sin(ANG10*leveltime)*step

			-- Skip and replace non-existent space graphics
			if not char:byte() or char:byte() == 32 then
				fx = $1+2*scale
				drawable = false
			end

			-- Unavoidable non V_ALLOWLOWERCASE flag toggle (exclude specials above 210)
			if (uppercs or (font == "CRFNT" or font == "NTFNT" or font == 'PTFNT'))
			and not (char:byte() >= 210) then
				char = tostring(char):upper()
			end
			-- transform the char to byte to a font patch
			local str = string.format("%s%03d", font, string.byte(char))
			charpatch = v.cachePatch(str)
			if not v.patchExists(str) then
				drawable = false
			end

			-- Draw char patch
			if drawable
				v.drawStretched(
					fx+off_x, fy+off_y+yscale,
					scale+hscale, scale+vscale, charpatch, flags, v.getColormap(nil,color))
				-- Sets the space between each character using font width
			end
			fx = $1+(xspacing+charpatch.width)*scale
			--fy = $1+yspacing*scale
		end

		-- Break new lines by spacing and patch width for semi-accurate spacing
		y = $1+(yspacing+charpatch.height)*scale >> FRACBITS 
	end	

end


rawset(_G, "drawSuperText", drawSuperText)
rawset(_G, "GetInternalFontWidth", GetInternalFontWidth)

rawset(_G, "L_TableContains", function(table, val)
	local hasit = {}
	for _,i in pairs(table) do
		hasit[i] = true
	end
	return hasit[val]
end)